
namespace cpse {
namespace math {

#define CPSE_MATRIX_OPERATOR_FLOAT(name, op) \
	template<class A, int R, int C> inline \
	tmatrix<matrix_binop<binop_##name, A, float_keeper> > \
	op(const tmatrix_dim<R, C, A> &a, float b) \
		{ return tmatrix<matrix_binop<binop_##name, A, \
			float_keeper> >(a, b); }

#define CPSE_FLOAT_OPERATOR_MATRIX(name, op) \
	template<class A, int R, int C> inline \
	tmatrix<matrix_binop<binop_##name, float_keeper, A> > \
	op(float a, const tmatrix_dim<R, C, A> &b) \
		{ return tmatrix<matrix_binop<binop_##name, \
			float_keeper, A> >(a, b); }

#define CPSE_MATRIX_UNARY_OPERATOR(name, op) \
	template<class A, int R, int C> inline \
	tmatrix<matrix_unop<unop_##name, A> > \
	op(const tmatrix_dim<R, C, A> &a) \
		{ return tmatrix<matrix_unop<unop_##name, A> >(a); }

#define CPSE_MATRIX_OPERATOR_MATRIX_ADD(name, op) \
	template<class A, class B, int R, int C> inline \
	tmatrix<matrix_binop<binop_##name, A, B> > \
	op(const tmatrix_dim<R, C, A> &a, const tmatrix_dim<R, C, B> &b) \
		{ return tmatrix<matrix_binop<binop_##name, A, B> >(a, b); }

// matrix multiplication
template<class A, class B, int C, int D, int E> inline
tmatrix<matrix_binop<binop_mul, A, B> >
operator*(const tmatrix_dim<C, D, A> &a, const tmatrix_dim<D, E, B> &b)
	{ return tmatrix<matrix_binop<binop_mul, A, B> >(a, b); }

// scalar product
template<class A, class B, int R> inline
float operator*(const tmatrix_dim<R, 1, A> &a, const tmatrix_dim<R, 1, B> &b)
	{ return dot(a, b); }

// scalar or vector product
template<class A, class B> inline
tmatrix<matrix_binop<binop_mul, A, B> >
operator*(const tmatrix_dim<3, 1, A> &a, const tmatrix_dim<3, 1, B> &b)
	{ return tmatrix<matrix_binop<binop_mul, A, B> >(a, b); }

CPSE_FLOAT_OPERATOR_MATRIX(mul, operator*)
CPSE_MATRIX_OPERATOR_FLOAT(mul, operator*)
CPSE_MATRIX_OPERATOR_MATRIX_ADD(add, operator+)
CPSE_MATRIX_OPERATOR_MATRIX_ADD(sub, operator-)
CPSE_MATRIX_OPERATOR_FLOAT(div, operator/)

template<class A, class B, int R> inline
tmatrix<matrix_binop<binop_cross, A, B> >
cross(const tmatrix_dim<R, 1, A> &a, const tmatrix_dim<R, 1, B> &b)
	{ return tmatrix<matrix_binop<binop_cross, A, B> >(a, b); }

CPSE_MATRIX_UNARY_OPERATOR(negate, operator-)
CPSE_MATRIX_UNARY_OPERATOR(transpose, transpose)

// unary +
template<class A, int R, int C> inline
const tmatrix<A> &operator+(const tmatrix_dim<R, C, A> &a)
	{ return a; }

// make a translation matrix
template<class A> inline
tmatrix<matrix_unop<unop_translation, A> >
translation(const tmatrix_dim<3, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_translation, A> >(a); }

// make a scaling matrix
template<class A> inline
tmatrix<matrix_unop<unop_scaling, A> >
scaling(const tmatrix_dim<3, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_scaling, A> >(a); }
inline tmatrix<matrix_unop<unop_scaling, float_keeper> > scaling(float a)
	{ return tmatrix<matrix_unop<unop_scaling, float_keeper> >(a); }

// make a rotation matrix
template<class A> inline
tmatrix<matrix_binop<binop_rotation, A, float_keeper> >
rotation(const tmatrix_dim<3, 1, A> &axis, float radians) {
	return tmatrix<matrix_binop<binop_rotation, A, float_keeper> >
		(axis, radians);
}

// make a perspective projection matrix
matrix<4> frustum(float left, float right, float bottom, float top,
	float near, float far);

// make an orthographic projection matrix
matrix<4> ortho(float left, float right, float bottom, float top,
	float near, float far);

// inverse
template<class A, int RC> inline
tmatrix<matrix_unop<unop_inverse, A> > inverse(const tmatrix_dim<RC, RC, A> &a)
	{ return tmatrix<matrix_unop<unop_inverse, A> >(a); }

// adjoint
template<class A, int RC> inline
tmatrix<matrix_unop<unop_adjoint, A> > adjoint(const tmatrix_dim<RC, RC, A> &a)
	{ return tmatrix<matrix_unop<unop_adjoint, A> >(a); }

// part of matrix
template<class A, int R, int C> inline
tmatrix_ndim<matrix_unop<unop_slice, A> > slice(const tmatrix_dim<R, C, A> &a)
	{ return tmatrix_ndim<matrix_unop<unop_slice, A> >(a); }

// perspective division
template<class A> inline
tmatrix<matrix_unop<unop_wdiv, A> > wdiv(const tmatrix_dim<4, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_wdiv, A> >(a); }

// normalize vector
template<int R, class A> inline
tmatrix<matrix_unop<unop_normalize, A> >
normalize(const tmatrix_dim<R, 1, A> &a)
	{ return tmatrix<matrix_unop<unop_normalize, A> >(a); }

// expand vector
template<class A, int R> inline
tmatrix<matrix_binop<binop_expand, A, float_keeper> >
expand(const tmatrix_dim<R, 1, A> &a, float b)
	{ return tmatrix<matrix_binop<binop_expand, A, float_keeper> >(a, b); }

// remove row R and column C
template<int R, int C, class A, int D, int E> inline
tmatrix<matrix_unop<unop_cofactor<R, C>, A> >
cofactor(const tmatrix_dim<D, E, A> &a)
	{ return tmatrix<matrix_unop<unop_cofactor<R, C>, A> >(a); }

// same as above, usable from impementations of other operations
template<int R, int C, int A, int B> inline
tmatrix<matrix_unop<unop_cofactor<R, C>, basic_matrix<A, B> > >
cofactor(const basic_matrix<A, B> &a)
{
	return tmatrix<matrix_unop<unop_cofactor<R, C>,
		basic_matrix<A, B> > >(a);
}

// evaluate to temporary
template<int R, int C, class T> inline
matrix<R, C> temp(const tmatrix_dim<R, C, T> &a) { return a; }

// determinant
inline float det(float a1, float a2, float b1, float b2)
	{ return a1*b2 - a2*b1; }
inline float det(basic_matrix<2, 1> a, basic_matrix<2, 1> b)
	{ return a.x*b.y - a.y*b.x; }
inline float det_bm(const basic_matrix<1, 1> &m) { return m(0,0); }
inline float det_bm(const basic_matrix<2, 2> &m)
	{ return m(0,0)*m(1,1) - m(1,0)*m(0,1); }
inline float det(float a1, float a2, float a3, float b1, float b2, float b3, 
	float c1, float c2, float c3)
{
	return a1*b2*c3 + a2*b3*c1 + a3*b1*c2 - a3*b2*c1 - a1*b3*c2 - a2*b1*c3;
}
inline float det(basic_matrix<3, 1> a, basic_matrix<3, 1> b,
	basic_matrix<3, 1> c)
{
	return det(a.x, a.y, a.z, b.x, b.y, b.z, c.x, c.y, c.z);
}
float det_bm(const basic_matrix<3, 3> &);
float det_bm(const basic_matrix<4, 4> &);

// dispatch determinant
template<int A> struct select_determinant {
	static float det(const basic_matrix<A, A> &m)
		{ return det_bm(m); }
};
template<int RC, class A> inline float det(const tmatrix_dim<RC, RC, A> &a)
	{ return select_determinant<RC>::det(a); }

// explicit scalar product
inline float dot_bm(const basic_matrix<2, 1> &a, const basic_matrix<2, 1> &b)
	{ return a.x*b.x + a.y*b.y; }
inline float dot_bm(const basic_matrix<3, 1> &a, const basic_matrix<3, 1> &b)
	{ return a.x*b.x + a.y*b.y + a.z*b.z; }
inline float dot_bm(const basic_matrix<4, 1> &a, const basic_matrix<4, 1> &b)
	{ return a.x*b.x + a.y*b.y + a.z*b.z + a.w*b.w; }

// dispatch scalar product
template<int A> struct select_scalar_product {
	static inline float dot(const basic_matrix<A, 1> &a,
		const basic_matrix<A, 1> &b)
		{ return dot_bm(a, b); }
};
template<int R, class A, class B> inline
float dot(const tmatrix_dim<R, 1, A> &a, const tmatrix_dim<R, 1, B> &b)
	{ return select_scalar_product<R>::dot(a, b); }

// squared length of vector
inline float length_squared_bm(const basic_matrix<2, 1> &a)
	{ return a.x*a.x + a.y*a.y; }
inline float length_squared_bm(const basic_matrix<3, 1> &a)
	{ return a.x*a.x + a.y*a.y + a.z*a.z; }
inline float length_squared_bm(const basic_matrix<4, 1> &a)
	{ return a.x*a.x + a.y*a.y + a.z*a.z + a.w+a.w; }

// dispatch length_squared
template<int A> struct select_length_squared {
	static inline float lsq(const basic_matrix<A, 1> &a)
		{ return length_squared_bm(a); }
};
template<int R, class A> inline
float length_squared(const tmatrix_dim<R, 1, A> &a)
	{ return select_length_squared<R>::lsq(a); }

// length of vector
template<int R, class A> inline
float length(const tmatrix_dim<R, 1, A> &a)
	{ return std::sqrt(length_squared(a)); }

template<int R, int C, int I>
struct matrix_assigner {
	basic_matrix<R, C> &dest;
	matrix_assigner(basic_matrix<R, (I > R*C ? -99 : C)> &d) : dest(d) {}
	matrix_assigner<R, C, I+1> operator,(float f)
		{ dest(I/C, I%C) = f; return dest; }
};

// assign to matrix
template<int R, int C> inline
matrix_assigner<R, C, 1> operator<<(basic_matrix<R, C> &dest, float f)
	{ dest(0,0) = f; return dest; }

// compare matrices
template<int A, int B, class T>
bool operator==(const basic_matrix_storage<A, B, T> &a,
	const basic_matrix_storage<A, B, T> &b)
{
	for (int i=0; i!=A*B; ++i)
		if (!close(a.data[i], b.data[i]))
			return false;
	return true;
}

// print matrix
template<int A, int B, class T>
std::ostream &operator<<(std::ostream &s,
	const basic_matrix_storage<A, B, T> &m)
{
	s << "\n(";
	for (int r=0; r<A-1; ++r) {
		s << '(';
		for (int c=0; c<B-1; ++c)
			s << m(r,c) << ',';
		s << m(r,B-1) << ")\n ";
	}
	s << '(';
	for (int c=0; c<B-1; ++c)
		s << m(A-1,c) << ',';
	return s << m(A-1,B-1) << "))";
}

// print vector
template<int A, class T>
std::ostream &operator<<(std::ostream &s,
	const basic_matrix_storage<A, 1, T> &m)
{
	s << "(";
	for (int i=0; i<A-1; ++i)
		s << m(i) << ',';
	return s << m(A-1) << ")t";
}

} // namespace math
} // namespace cpse
